@@ -1,5 +1,3 @@
-# coding: utf-8
-
 """
 A MANAGESIEVE client.
 
@@ -9,16 +7,11 @@ a user to syntactically flawed scripts.
 
 Implementation based on RFC 5804.
 """
-from __future__ import print_function
-
 import base64
 import re
 import socket
 import ssl
 
-from future.utils import python_2_unicode_compatible
-import six
-
 from .digest_md5 import DigestMD5
 from . import tools
 
@@ -36,7 +29,6 @@ class Error(Exception):
     pass
 
 
-@python_2_unicode_compatible
 class Response(Exception):
     def __init__(self, code, data):
         self.code = code
@@ -46,7 +38,6 @@ class Response(Exception):
         return "%s %s" % (self.code, self.data)
 
 
-@python_2_unicode_compatible
 class Literal(Exception):
     def __init__(self, value):
         self.value = value
@@ -227,7 +218,7 @@ class Client(object):
         """
         ret = []
         for a in args:
-            if isinstance(a, six.binary_type):
+            if isinstance(a, bytes):
                 if self.__size_expr.match(a):
                     ret += [a]
                 else:
@@ -236,6 +227,19 @@ class Client(object):
             ret += [bytes(str(a).encode("utf-8"))]
         return ret
 
+    def __prepare_content(self, content):
+        """Format script content before sending it.
+
+        Script length must be inserted before the content,
+        enclosed in curly braces, separated by CRLF.
+
+        :param content: script content as str or bytes
+        :return: transformed script as bytes
+        """
+        if isinstance(content, str):
+            content = content.encode('utf-8')
+        return b"{%d+}%s%s" % (len(content), CRLF, content)
+
     def __send_command(
             self, name, args=None, withcontent=False, extralines=None,
             nblines=-1):
@@ -268,9 +272,10 @@ class Client(object):
                 self.sock.sendall(l + CRLF)
         code, data, content = self.__read_response(nblines)
 
-        if isinstance(code, six.binary_type):
+        if isinstance(code, bytes):
             code = code.decode("utf-8")
-        data = data.decode("utf-8")
+        if isinstance(data, bytes):
+            data = data.decode("utf-8")
 
         if withcontent:
             return (code, data, content)
@@ -327,9 +332,9 @@ class Client(object):
         :param password: clear password
         :return: True on success, False otherwise.
         """
-        if isinstance(login, six.text_type):
+        if isinstance(login, str):
             login = login.encode("utf-8")
-        if isinstance(password, six.text_type):
+        if isinstance(password, str):
             password = password.encode("utf-8")
         params = base64.b64encode(b'\0'.join([authz_id, login, password]))
         code, data = self.__send_command("AUTHENTICATE", [b"PLAIN", params])
@@ -477,7 +482,7 @@ class Client(object):
 
         :rtype: string
         """
-        if isinstance(self.__capabilities["SIEVE"], six.string_types):
+        if isinstance(self.__capabilities["SIEVE"], str):
             self.__capabilities["SIEVE"] = self.__capabilities["SIEVE"].split()
         return self.__capabilities["SIEVE"]
 
@@ -496,8 +501,7 @@ class Client(object):
         :rtype: boolean
         """
         try:
-            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            self.sock.connect((self.srvaddr, self.srvport))
+            self.sock = socket.create_connection((self.srvaddr, self.srvport))
             self.sock.settimeout(Client.read_timeout)
         except socket.error as msg:
             raise Error("Connection to server failed: %s" % str(msg))
@@ -605,8 +609,7 @@ class Client(object):
         :param content: script's content
         :rtype: boolean
         """
-        content = tools.to_bytes(content)
-        content = tools.to_bytes("{%d+}" % len(content)) + CRLF + content
+        content = self.__prepare_content(content)
         code, data = (
             self.__send_command("PUTSCRIPT", [name.encode("utf-8"), content]))
         if code == "OK":
@@ -702,8 +705,7 @@ class Client(object):
         if "VERSION" not in self.__capabilities:
             raise NotImplementedError(
                 "server does not support CHECKSCRIPT command")
-        content = tools.to_bytes(
-            u"{%d+}%s%s" % (len(content), str(CRLF), content))
+        content = self.__prepare_content(content)
         code, data = self.__send_command("CHECKSCRIPT", [content])
         if code == "OK":
             return True
