@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# coding: utf-8
 
 """
 This module provides a simple but functional parser for the SIEVE
@@ -8,19 +7,13 @@ language used to filter emails.
 This implementation is based on RFC 5228 (http://tools.ietf.org/html/rfc5228)
 
 """
-from __future__ import print_function
-
 import re
 import sys
 
-from future.utils import python_2_unicode_compatible, text_type
-import six
-
 from sievelib.commands import (
     get_command_instance, CommandError, RequireCommand)
 
 
-@python_2_unicode_compatible
 class ParseError(Exception):
     """Generic parsing error"""
 
@@ -48,9 +41,7 @@ class Lexer(object):
         self.definitions = definitions
         parts = []
         for name, part in definitions:
-            param = "(?P<%s>%s)" % (name.decode(), part.decode())
-            if six.PY3:
-                param = bytes(param, "utf-8")
+            param = b"(?P<%s>%s)" % (name, part)
             parts.append(param)
         self.regexpString = b"|".join(parts)
         self.regexp = re.compile(self.regexpString, re.MULTILINE)
@@ -60,6 +51,10 @@ class Lexer(object):
         """Return the current line number"""
         return self.text[:self.pos].count(b'\n') + 1
 
+    def curcolno(self):
+        """Return the current column number"""
+        return self.pos - self.text.rfind(b'\n', 0, self.pos)
+
     def scan(self, text):
         """Analyse some data
 
@@ -81,10 +76,14 @@ class Lexer(object):
 
             m = self.regexp.match(text, self.pos)
             if m is None:
-                raise ParseError("unknown token %s" % text[self.pos:])
+                token = text[self.pos:]
+                m = self.wsregexp.search(token)
+                if m is not None:
+                    token = token[:m.start()]
+                raise ParseError("unknown token %s" % token)
 
-            self.pos = m.end()
             yield (m.lastgroup, m.group(m.lastgroup))
+            self.pos += len(m.group(0))
 
 
 class Parser(object):
@@ -134,7 +133,7 @@ class Parser(object):
         self.__curcommand = None
         self.__curstringlist = None
         self.__expected = None
-        self.__opened_blocks = 0
+        self.__expected_brackets = []
         RequireCommand.loaded_extensions = []
 
     def __set_expected(self, *args, **kwargs):
@@ -145,6 +144,28 @@ class Parser(object):
         """
         self.__expected = args
 
+    def __push_expected_bracket(self, ttype, tvalue):
+        """Append a new expected bracket.
+
+        Next time a bracket is closed, it must match the one provided here.
+        """
+        self.__expected_brackets.append((ttype, tvalue))
+
+    def __pop_expected_bracket(self, ttype, tvalue):
+        """Drop the last expected bracket.
+
+        If the given bracket doesn't match the dropped expected bracket,
+        or if no bracket is expected at all, a ParseError will be raised.
+        """
+        try:
+            etype, evalue = self.__expected_brackets.pop()
+        except IndexError:
+            raise ParseError("unexpected closing bracket %s (none opened)" %
+                             (tvalue,))
+        if ttype != etype:
+            raise ParseError("unexpected closing bracket %s (expected %s)" %
+                             (tvalue, evalue))
+
     def __up(self, onlyrecord=False):
         """Return to the current command's parent
 
@@ -173,7 +194,17 @@ class Parser(object):
             self.result += [self.__curcommand]
 
         if not onlyrecord:
-            self.__curcommand = self.__curcommand.parent
+            while self.__curcommand:
+                self.__curcommand = self.__curcommand.parent
+                # Make sure to detect all done tests (including 'not' ones).
+                condition = (
+                    self.__curcommand and
+                    self.__curcommand.get_type() == "test" and
+                    self.__curcommand.iscomplete()
+                )
+                if condition:
+                    continue
+                break
 
     def __check_command_completion(self, testsemicolon=True):
         """Check for command(s) completion
@@ -233,6 +264,7 @@ class Parser(object):
             self.__set_expected("string")
             return True
         if ttype == "right_bracket":
+            self.__pop_expected_bracket(ttype, tvalue)
             self.__curcommand.check_next_arg("stringlist", self.__curstringlist)
             self.__cstate = self.__arguments
             return self.__check_command_completion()
@@ -257,10 +289,22 @@ class Parser(object):
             return self.__curcommand.check_next_arg(ttype, tvalue.decode("ascii"))
 
         if ttype == "left_bracket":
+            self.__push_expected_bracket("right_bracket", b'}')
             self.__cstate = self.__stringlist
             self.__curstringlist = []
             self.__set_expected("string")
             return True
+
+        condition = (
+            ttype in ["left_cbracket", "comma"] and
+            self.__curcommand.non_deterministic_args
+        )
+        if condition:
+            self.__curcommand.reassign_arguments()
+            # rewind lexer
+            self.lexer.pos -= 1
+            return True
+
         return False
 
     def __arguments(self, ttype, tvalue):
@@ -285,6 +329,7 @@ class Parser(object):
             return self.__check_command_completion(testsemicolon=False)
 
         if ttype == "left_parenthesis":
+            self.__push_expected_bracket("right_parenthesis", b')')
             self.__set_expected("identifier")
             return True
 
@@ -293,6 +338,7 @@ class Parser(object):
             return True
 
         if ttype == "right_parenthesis":
+            self.__pop_expected_bracket(ttype, tvalue)
             self.__up()
             return True
 
@@ -319,8 +365,8 @@ class Parser(object):
         """
         if self.__cstate is None:
             if ttype == "right_cbracket":
+                self.__pop_expected_bracket(ttype, tvalue)
                 self.__up()
-                self.__opened_blocks -= 1
                 self.__cstate = None
                 return True
 
@@ -347,7 +393,7 @@ class Parser(object):
             return True
 
         if ttype == "left_cbracket":
-            self.__opened_blocks += 1
+            self.__push_expected_bracket("right_cbracket", b'}')
             self.__cstate = None
             return True
 
@@ -375,11 +421,12 @@ class Parser(object):
         :param text: a string containing the data to parse
         :return: True on success (no error detected), False otherwise
         """
-        if isinstance(text, text_type):
+        if isinstance(text, str):
             text = text.encode("utf-8")
 
         self.__reset_parser()
         try:
+            tvalue = ''
             for ttype, tvalue in self.lexer.scan(text):
                 if ttype == "hash_comment":
                     self.hash_comments += [tvalue.strip()]
@@ -388,27 +435,35 @@ class Parser(object):
                     continue
                 if self.__expected is not None:
                     if ttype not in self.__expected:
-                        if self.lexer.pos < len(text):
-                            msg = "%s found while %s expected near '%s'" \
-                                  % (ttype, "|".join(self.__expected), text[self.lexer.pos])
+                        if self.lexer.pos < len(text) + len(tvalue):
+                            msg = (
+                                "%s found while %s expected near '%s'"
+                                % (ttype, "|".join(self.__expected),
+                                   text[self.lexer.pos])
+                            )
                         else:
-                            msg = "%s found while %s expected at end of file" \
+                            msg = (
+                                "%s found while %s expected at end of file"
                                   % (ttype, "|".join(self.__expected))
+                            )
                         raise ParseError(msg)
                     self.__expected = None
 
                 if not self.__command(ttype, tvalue):
-                    msg = "unexpected token '%s' found near '%s'" \
-                          % (tvalue, text[self.lexer.pos])
+                    msg = (
+                        "unexpected token '%s' found near '%s'"
+                        % (tvalue.decode(), text.decode()[self.lexer.pos])
+                    )
                     raise ParseError(msg)
-            if self.__opened_blocks:
-                self.__set_expected("right_cbracket")
+            if self.__expected_brackets:
+                self.__set_expected(self.__expected_brackets[-1][0])
             if self.__expected is not None:
                 raise ParseError("end of script reached while %s expected" %
                                  "|".join(self.__expected))
 
         except (ParseError, CommandError) as e:
-            self.error = "line %d: %s" % (self.lexer.curlineno(), str(e))
+            self.error_pos = (self.lexer.curlineno(), self.lexer.curcolno(), len(tvalue))
+            self.error = "line %d: %s" % (self.error_pos[0], str(e))
             return False
         return True
 
