@@ -1,12 +1,10 @@
-# coding: utf-8
-
 """
 Unit tests for the SIEVE language parser.
 """
 import unittest
 import os.path
 import codecs
-import six
+import io
 
 from sievelib.parser import Parser
 from sievelib.factory import FiltersSet
@@ -52,14 +50,14 @@ class SieveTest(unittest.TestCase):
     def __checkCompilation(self, script, result):
         self.assertEqual(self.parser.parse(script), result)
 
-    def compilation_ok(self, script):
-        self.__checkCompilation(script, True)
+    def compilation_ok(self, script, **kwargs):
+        self.__checkCompilation(script, True, **kwargs)
 
     def compilation_ko(self, script):
         self.__checkCompilation(script, False)
 
     def representation_is(self, content):
-        target = six.StringIO()
+        target = io.StringIO()
         self.parser.dump(target)
         repr_ = target.getvalue()
         target.close()
@@ -68,7 +66,7 @@ class SieveTest(unittest.TestCase):
     def sieve_is(self, content):
         filtersset = FiltersSet("Testfilterset")
         filtersset.from_parser_result(self.parser)
-        target = six.StringIO()
+        target = io.StringIO()
         filtersset.tosieve(target)
         repr_ = target.getvalue()
         target.close()
@@ -189,7 +187,7 @@ noreply
 Your email has been canceled
 ============================
 .
-    stop (type: control)
+    stop (type: action)
 else (type: control)
     reject (type: action)
         text:
@@ -199,6 +197,41 @@ Your email has been canceled too
 .
 """)
 
+    def test_complex_allof_with_not(self):
+        """Test for allof/anyof commands including a not test.
+
+        See https://github.com/tonioo/sievelib/issues/69.
+        """
+        self.compilation_ok(b"""
+require ["fileinto", "reject"];
+
+if allof (not allof (address :is ["From","sender"] ["test1@test2.priv","test2@test2.priv"], header :matches "Subject" "INACTIVE*"), address :is "From" "user3@test3.priv")
+{
+    reject;
+}
+""")
+        self.representation_is("""
+require (type: control)
+    ["fileinto","reject"]
+if (type: control)
+    allof (type: test)
+        not (type: test)
+            allof (type: test)
+                address (type: test)
+                    :is
+                    ["From","sender"]
+                    ["test1@test2.priv","test2@test2.priv"]
+                header (type: test)
+                    :matches
+                    "Subject"
+                    "INACTIVE*"
+        address (type: test)
+            :is
+            "From"
+            "user3@test3.priv"
+    reject (type: action)
+""")
+
     def test_nested_blocks(self):
         self.compilation_ok(b"""
 if header :contains "Sender" "example.com" {
@@ -259,7 +292,7 @@ if header :is "Sender" "owner-ietf-mta-filters@imc.org"
 #
 # Keep all messages to or from people in my company
 #
-elsif address :DOMAIN :is ["From", "To"] "example.com"
+elsif address :domain :is ["From", "To"] "example.com"
         {
         keep;               # keep in "In" mailbox
         }
@@ -294,7 +327,7 @@ if (type: control)
         "filter"
 elsif (type: control)
     address (type: test)
-        :DOMAIN
+        :domain
         :is
         ["From","To"]
         "example.com"
@@ -327,6 +360,7 @@ if header :contains :comparator "i;octet" "Subject" "MAKE MONEY FAST" {
         self.representation_is("""
 if (type: control)
     header (type: test)
+        :comparator
         "i;octet"
         :contains
         "Subject"
@@ -363,7 +397,7 @@ if (type: control)
             not (type: test)
                 not (type: test)
                     true (type: test)
-    stop (type: control)
+    stop (type: action)
 """)
 
     def test_just_one_command(self):
@@ -471,6 +505,66 @@ if header :contains "subject" "viagra" {
 }
 """)
 
+    def test_fileinto_create(self):
+        self.compilation_ok(b"""require ["fileinto", "mailbox"];
+if header :is "Sender" "owner-ietf-mta-filters@imc.org"
+        {
+        fileinto :create "filter";  # move to "filter" mailbox
+        }
+""")
+
+    def test_imap4flags_extension(self):
+        self.compilation_ok(b"""
+require ["fileinto", "imap4flags", "variables"];
+if size :over 1M {
+    addflag "MyFlags" "Big";
+    if header :is "From" "boss@company.example.com" {
+       # The message will be marked as "\Flagged Big" when filed into
+       # mailbox "Big messages"
+       addflag "MyFlags" "\\Flagged";
+    }
+    fileinto :flags "${MyFlags}" "Big messages";
+}
+""")
+
+    def test_imap4flags_hasflag(self):
+        self.compilation_ok(b"""
+require ["imap4flags", "fileinto"];
+
+if hasflag ["test", "toto"] {
+    fileinto "Test";
+}
+addflag "Var1" "Truc";
+if hasflag "Var1" "Truc" {
+    fileinto "Truc";
+}
+""")
+
+    def test_body_extension(self):
+        self.compilation_ok(b"""
+require ["body", "fileinto"];
+
+if body :content "text" :contains ["missile", "coordinates"] {
+    fileinto "secrets";
+}
+""")
+        self.compilation_ok(b"""
+require "body";
+
+if body :raw :contains "MAKE MONEY FAST" {
+    discard;
+}
+""")
+        self.compilation_ok(b"""
+require ["body", "fileinto"];
+
+# Save messages mentioning the project schedule in the
+# project/schedule folder.
+if body :text :contains "project schedule" {
+    fileinto "project/schedule";
+}
+""")
+
 
 class InvalidSyntaxes(SieveTest):
     def test_nested_comments(self):
@@ -481,7 +575,7 @@ it is allowed by the RFC :p */
 
     def test_nonopened_block(self):
         self.compilation_ko(b"""
-if header :is "Sender" "me@example.com" 
+if header :is "Sender" "me@example.com"
     discard;
 }
 """)
@@ -493,16 +587,29 @@ if header :is "Sender" "me@example.com" {
 
 """)
 
+    def test_nonopened_parenthesis(self):
+        self.compilation_ko(b"""
+if header :is "Sender" "me@example.com") {
+    discard;
+}
+""")
+
+    def test_nonopened_block2(self):
+        self.compilation_ko(b"""}""")
+
     def test_unknown_token(self):
         self.compilation_ko(b"""
 if header :is "Sender" "Toto" & header :contains "Cc" "Tata" {
-    
+
 }
 """)
 
     def test_empty_string_list(self):
         self.compilation_ko(b"require [];")
 
+    def test_unopened_string_list(self):
+        self.compilation_ko(b'require "fileinto"];')
+
     def test_unclosed_string_list(self):
         self.compilation_ko(b'require ["toto", "tata";')
 
@@ -648,12 +755,56 @@ if header :contains "Subject" "MAKE MONEY FAST" {
     def test_test_outside_control(self):
         self.compilation_ko(b"true;")
 
+    def test_fileinto_create_without_mailbox(self):
+        self.compilation_ko(b"""require ["fileinto"];
+if header :is "Sender" "owner-ietf-mta-filters@imc.org"
+        {
+        fileinto :create "filter";  # move to "filter" mailbox
+        }
+""")
+        self.assertEqual(
+            self.parser.error, "line 4: extension 'mailbox' not loaded")
+
+    def test_fileinto_create_without_fileinto(self):
+        self.compilation_ko(b"""require ["mailbox"];
+if header :is "Sender" "owner-ietf-mta-filters@imc.org"
+        {
+        fileinto :create "filter";  # move to "filter" mailbox
+        }
+""")
+        self.assertEqual(
+            self.parser.error, "line 4: unknown command fileinto")
+
+    def test_exists_get_string_or_list(self):
+        self.compilation_ok(b"""
+if exists "subject"
+{
+       discard;
+}
+""")
+        self.compilation_ok(b"""
+if exists ["subject"]
+{
+       discard;
+}
+""")
+
 
 class DateCommands(SieveTest):
+
+    def test_date_command(self):
+        self.compilation_ok(b"""require ["date", "relational", "fileinto"];
+if allof(header :is "from" "boss@example.com",
+         date :value "ge" :originalzone "date" "hour" "09",
+         date :value "lt" :originalzone "date" "hour" "17")
+{ fileinto "urgent"; }
+""")
+
     def test_currentdate_command(self):
         self.compilation_ok(b"""require ["date", "relational"];
 
-if allof ( currentdate :value "ge" "date" "2013-10-23" , currentdate :value "le" "date" "2014-10-12" ) 
+if allof(currentdate :value "ge" "date" "2013-10-23",
+         currentdate :value "le" "date" "2014-10-12")
 {
     discard;
 }
@@ -662,7 +813,8 @@ if allof ( currentdate :value "ge" "date" "2013-10-23" , currentdate :value "le"
     def test_currentdate_command_timezone(self):
         self.compilation_ok(b"""require ["date", "relational"];
 
-if allof ( currentdate :zone "+0100" :value "ge" "date" "2013-10-23" , currentdate :value "le" "date" "2014-10-12" ) 
+if allof(currentdate :zone "+0100" :value "ge" "date" "2013-10-23",
+         currentdate :value "le" "date" "2014-10-12")
 {
     discard;
 }
@@ -671,20 +823,29 @@ if allof ( currentdate :zone "+0100" :value "ge" "date" "2013-10-23" , currentda
     def test_currentdate_norel(self):
         self.compilation_ok(b"""require ["date"];
 
-if allof ( 
-  currentdate :zone "+0100" :is "date" "2013-10-23"  
-) 
+if allof (
+  currentdate :zone "+0100" :is "date" "2013-10-23"
+)
 {
     discard;
 }""")
 
+    def test_currentdate_extension_not_loaded(self):
+        self.compilation_ko(b"""require ["date"];
+
+if allof ( currentdate :value "ge" "date" "2013-10-23" , currentdate :value "le" "date" "2014-10-12" )
+{
+    discard;
+}
+""")
+
 
 class VariablesCommands(SieveTest):
     def test_set_command(self):
         self.compilation_ok(b"""require ["variables"];
 
 set "matchsub" "testsubject";
-        
+
 if allof (
   header :contains ["Subject"] "${header}"
 )
